#include "../std.h"

extern "C" {
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
#include <globus_replica_catalog.h>
#endif
}

#include "../config/config_map.h"
#include "../misc/url_options.h"
#include "../misc/checksum.h"
#include "../misc/globus_modules.h"
#include "../misc/log_time.h"

#include "datapoint.h"

DataPoint::~DataPoint(void) {
}

DataPoint::DataPoint(const char* url_) {
  is_valid=false; tries_left=5;
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  if((!common_mod.active()) || (!io_mod.active()) || (!rls_mod.active())) return;
#endif
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
  if(!replica_mod.active()) return;
#endif
  if((!error_mod.active()) || (!gsi_gssapi_mod.active())) return;
  if((url_ == NULL) || (url_[0]==0)) return;
  common_url_options="";
  is_valid=true; url=url_;
  locations.clear(); location=locations.end();
  is_stdchannel=false; is_file=false; is_ftp=false; is_http=false;
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
  is_rc=false; 
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  is_rls=false;
#endif
#ifndef SKIP_HTTPG
  is_httpg=false; is_se=false;
#endif
/* is_magda=false; is_bbftp=false; */
  is_secure=false; is_resolved=true; 
  meta_size_valid=false; meta_checksum_valid=false;
  meta_created_valid=false; meta_validtill_valid=false;
  is_metaexisting=false;
/*  meta_valid=false; */
  if(strcmp("-",url_) == 0) {
    is_stdchannel=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("file:/",url_,6) == 0) {
    is_file=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
#ifdef HAVE_GLOBUS_REPLICA_CATALOG_H
  else if(strncasecmp("rc://",url_,5) == 0) {
    is_rc=true; is_secure=true; is_resolved=false;
    if(!process_meta_url_rc(url_)) {
      is_valid=false;
    };
    location=locations.begin(); rc_mgr=NULL;
  }
#endif
#ifdef HAVE_GLOBUS_RLS_CLIENT_H
  else if(strncasecmp("rls://",url_,6) == 0) {
    is_rls=true; is_secure=true; is_resolved=false;
    if(!process_meta_url_rls(url_)) {
      is_valid=false;
    };
    location=locations.begin();
  }
#endif
  else if(strncasecmp("ftp://",url_,6) == 0) {
    is_ftp=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("gsiftp://",url_,9) == 0) {
    is_ftp=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("http://",url_,7) == 0) {
    is_http=true; is_secure=false;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("https://",url_,8) == 0) {
    is_http=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
#ifndef SKIP_HTTPG
  else if(strncasecmp("httpg://",url_,8) == 0) {
    is_httpg=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
  else if(strncasecmp("se://",url_,5) == 0) {
    is_se=true; is_secure=true;
    locations.push_back(url_); location=locations.begin();
  }
#endif
  else {
    odlog(0)<<"DataPoint: unrecognized protocol"<<endl;
    locations.push_back(url_); // to avoid segfauls if is_valid not checked
    is_valid=false; location=locations.begin();
  };
}

bool DataPoint::get_info(DataPoint::FileInfo &fi) {
  if(!meta()) { return true; }; // not a meta url
  if(!meta_resolve(true)) { return false; }; // resolution failure
  fi.name=lfn();
  for(list<DataPoint::Location>::iterator i = locations.begin();
                                              i!=locations.end();++i) {
    fi.urls.push_back(i->url);
  };  
  if(meta_size_valid) { fi.size=meta_size_; fi.size_available=true; };
  if(meta_checksum_valid) { fi.checksum=meta_checksum_; fi.checksum_available=true; };
  if(meta_created_valid) { fi.created=meta_created_; fi.created_available=true; };
  if(meta_validtill_valid) { fi.valid=meta_validtill_; fi.valid_available=true; };
  fi.type=DataPoint::FileInfo::file_type_file;
  return true;
}

bool DataPoint::list_files(list<DataPoint::FileInfo> &files,bool resolve) {
  if(!meta()) return true;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"list: RC: url: "<<url<<endl;
    odlog(2)<<"list: RC: service url: "<<meta_service_url<<endl;
    odlog(2)<<"list: RC: lfn: "<<meta_lfn<<endl;
    return list_files_rc(files,resolve);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"list: RLS: url: "<<url<<endl;
    odlog(2)<<"list: RLS: service url: "<<meta_service_url<<endl;
    odlog(2)<<"list: RLS: lfn: "<<meta_lfn<<endl;
    return list_files_rls(files,resolve);
  };
  return false;
}

/* perform resolve operation, which can take long time */
bool DataPoint::meta_resolve(bool source,const UrlMap &maps) {
  if(is_resolved) return true;
  bool res=meta_resolve(source);
  if(!res) return false;
  sort(maps); location=locations.begin();
  return true;
}

bool DataPoint::meta_resolve(bool source) {
  if(is_resolved) return true;
  if(!meta()) return true;
  odlog(2)<<"meta_resolve: url: "<<url<<endl;
  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_resolve: RC url"<<endl;
    return meta_resolve_rc(source);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_resolve: RLS url"<<endl;
    return meta_resolve_rls(source);
  };
  return false;
}

bool DataPoint::meta_preregister(bool replication,bool force) {
  if(!meta()) return true;
  odlog(2)<<"meta_preregister"<<endl;

  if((location != locations.end()) &&
     (strncasecmp(location->url.c_str(),"se://",5) == 0)) {
    return true; // SE registers files itself
  };

  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    return meta_preregister_rc(replication,force);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    return meta_preregister_rls(replication,force);
  };
  return true;
}

bool DataPoint::meta_postregister(bool replication,bool failure) {
  if(!meta()) return true;
  odlog(2)<<"meta_postregister"<<endl;

  if((location != locations.end()) &&
     (strncasecmp(location->url.c_str(),"se://",5) == 0)) {
    return true; // SE registers files itself
  };

  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_postregister: is rc"<<endl;
    return meta_postregister_rc(replication,failure);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_postregister: is rls"<<endl;
    return meta_postregister_rls(replication,failure);
  };
  return true;
}

bool DataPoint::meta_unregister(bool all) {
  if(!meta()) return true;

  if((!all) &&
     (location != locations.end()) &&
     (strncasecmp(location->url.c_str(),"se://",5) == 0)) {
    return true; // SE registers files itself
  };

  if(strncasecmp("rc://",url.c_str(),5) == 0) {
/* not fully implemented yet - all=true will remove logical file only */
    odlog(2)<<"meta_unregister: is rc"<<endl;
    return meta_unregister_rc(all);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_unregister: is rls"<<endl;
    return meta_unregister_rls(all);
  };
  return true;
}

bool DataPoint::meta_preunregister(bool replication) {
  if(!meta()) return true;

  if((location != locations.end()) &&
     (strncasecmp(location->url.c_str(),"se://",5) == 0)) {
    return true; // SE registers files itself
  };

  if(strncasecmp("rc://",url.c_str(),5) == 0) {
    odlog(2)<<"meta_preunregister: is rc"<<endl;
    return meta_preunregister_rc(replication);
  }
  else if(strncasecmp("rls://",url.c_str(),6) == 0) {
    odlog(2)<<"meta_preunregister: is rls"<<endl;
    return meta_preunregister_rls(replication);
  };
  return true;
}

bool DataPoint::have_locations(void) {
  if(!is_valid) return false;
  return (locations.size() != 0);
}

bool DataPoint::have_location(void) {
  if(!is_valid) return false;
  if(!is_resolved) {
    if(tries_left <= 0) return false;
    return true;
  };
  if(location == locations.end()) return false;
  return true;
}

bool DataPoint::next_location(void) {
  if(!is_resolved) {
    tries_left--;
    if(tries_left <= 0) return false;
    return true;
  };
  if(location == locations.end()) return false;
  ++location;
  if(location == locations.end()) {
    tries_left--;
    if(tries_left <= 0) return false;
    location=locations.begin();
  };
  return true;
}

bool DataPoint::remove_location(void) {
  if(location == locations.end()) return false;
  location=locations.erase(location);
  return true;
}

bool DataPoint::remove_locations(const DataPoint& p) {
  list<Location>::iterator p_int;
  list<Location>::const_iterator p_ext;
  for(p_ext=p.locations.begin();p_ext!=p.locations.end();++p_ext) {
    for(p_int=locations.begin();p_int!=locations.end();) {
        if(p_int->url == p_ext->url) {
        if(location == p_int) {
          p_int=locations.erase(p_int); location=p_int;
        } else {
          p_int=locations.erase(p_int);
        };
        continue;
      };
      ++p_int;
    };
  };
  if(location == locations.end()) location=locations.begin();
  return true;
}

string DataPoint::canonic_url(void) const {
  string tmp = url;
  if(tmp == "-") { return tmp; };
  if(::canonic_url(tmp) != 0) tmp="";
  return tmp;
}

string DataPoint::base_url(void) const {
  return url;
}

ostream& operator<<(ostream& o,const DataPoint &point) {
  if(!point.is_valid) { o<<"<invalid>"; return o; };
  if(point.meta()) {
    if(point.is_resolved) {
      if(point.locations.end() != point.location) {
        o<<point.url<<"["<<point.location->url<<"]";
        return o;
      };
    };   
  };  
  o<<point.url;
  return o;
}

/* put local urls first and shuffle rest */
bool DataPoint::sort(const UrlMap &maps) {
  list<Location>::iterator ii = locations.begin();
  int nn = 0;
  for(list<Location>::iterator i=locations.begin();i!=locations.end();) {
    string c_url = i->url;
    ::canonic_url(c_url);
    if(maps.local(c_url)) {  /* move to first place */
      if(i == ii) { ++ii; ++i; nn++; continue; };
      locations.insert(ii,*i);
      if(i == location) location=locations.begin();
      i=locations.erase(i);
      nn++;
    }
    else {
      ++i;
    };
  };
  nn=locations.size()-nn;
  if(nn<=1) return true;
  srandom(time(NULL));
  for(;nn>1;) {
    int r = random()/2;
    int n=r/((RAND_MAX/2+1)/nn);
    list<Location>::iterator i=ii;
    for(;n;n--) { ++i; };
    if(i == locations.end()) { nn--; ++ii; continue; };
    if(i == ii) { nn--; ++ii; continue; };
    locations.insert(ii,*i);
    if(i == location) location=locations.begin();
    i=locations.erase(i);
    nn--;
  };
  return true;
}

bool DataPoint::map(const UrlMap &maps) {
  for(list<Location>::iterator i=locations.begin();i!=locations.end();) {
    if(maps.map(i->url)) {  /* move to first place */
      locations.push_front(*i);
      if(i == location) location=locations.begin();
      i=locations.erase(i);
    }
    else {
      ++i;
    };
  };
  return true;
}

bool DataPoint::add_location(const char* meta_loc,const char* loc) {
  if(!meta()) return false;
  odlog(2)<<"add_location: meta location: "<<meta_loc<<endl;
  odlog(2)<<"add_location: location: "<<loc<<endl;
  for(list<Location>::iterator i = locations.begin();i!=locations.end();++i) {
    if(i->meta == meta_loc) return true;
  };
  locations.insert(locations.end(),Location(meta_loc,loc,false));
  return true;
}

bool DataPoint::extract_meta_attributes(string& lfn) {
  meta_attributes.clear();
  int attribute_start = lfn.find(':',0);
  if( attribute_start == string::npos ) return true;
  string allattributes = lfn.substr( attribute_start+1 );
  lfn.erase( attribute_start );
  attribute_start = 0;
  int new_attribute_start;
  do {
    new_attribute_start = allattributes.find(':',attribute_start);
    string attribute = allattributes.substr( attribute_start, 
      new_attribute_start!=string::npos?new_attribute_start-attribute_start:string::npos
    );
    int findvalue = attribute.find('=');
    if( findvalue != string::npos ) {
      string value = attribute.substr( findvalue+1 );
      string name = attribute.substr( 0, findvalue );
      odlog(2) << "Attribute: "<< name << " = " << value << endl;
      meta_attributes[ name ] = value;
    } else {
      odlog(2) << "Invalid attribute: " << attribute << endl;
    }
    attribute_start=new_attribute_start+1;
  } while( new_attribute_start != string::npos );
  return true;
}

